import time
import math
import numpy as np
import serial
import matplotlib
matplotlib.use('TKAgg')
from math import pi
from matplotlib import pyplot as plt

plt.ion()


class ZMapGenerator(object):
    def __init__(
            self,
            speed_max_travel=15000,
            speed_max_dot=9000,
            wait_time_home=5,
            wait_time_xyz_zero=5,
            wait_time_point=0.3,
            wait_time_row=5,
            z_height_row_change=15,
            r=65,
            step_size=10,
            csv_file="zmap.csv",
            com_port="/dev/ttyACM0",
            baud_rate=115200,
            demo=False):

        self.r = r

        if demo:
            self.mf = self.get_demo_measurement
        else:
            self.mf = self.get_measurement_from_user

        self.travel = speed_max_travel
        self.point_speed = speed_max_dot
        self.wth = wait_time_home
        self.wtz = wait_time_xyz_zero
        self.wtp = wait_time_point
        self.wtr = wait_time_row
        self.step_size = step_size
        self.z_height_row_change = z_height_row_change
        self.points = []
        self.points_adjusted = []
        self.csv_file = csv_file
        self.printer = None
        self.writer = None
        self.com_port = com_port
        self.baud_rate = baud_rate
        self.figure, self.axes = plt.subplots(1, 1)
        self.axes.set_aspect('equal')
        self.axes.set_xlim(-r-5, r+5)
        self.axes.set_ylim(-r-5, r+5)
        self.axes.hold(True)
        plt.show(False)
        plt.draw()
        self.bbox = self.axes.bbox
        self.background = self.figure.canvas.copy_from_bbox(self.bbox)
        self.main()

    def get_measurement_from_user(self, x, y):
        while True:
            try:
                measurement = float(raw_input("[{0},{1}]: Enter Z Measurement: ".format(x, y)).strip())
                break
            except:
                pass
        return measurement

    def get_demo_measurement(self, x, y):
        return np.random.random()

    def connect_to_printer(self):
        try:
            self.printer = serial.Serial(self.com_port, self.baud_rate)
        except Exception, e:
            print "Failed to connect to printer =( -- ", e
            self.writer = self.write_to_stdout
            return

        wait_count = 0
        while True:
            data = self.printer.readline()
            print "From {0}: {1}".format(self.com_port, data)

            if "wait" in data:
                wait_count += 1

            if wait_count == 10:
                break

        print "Connected to printer on port {0}".format(self.com_port)
        self.writer = self.write_to_printer

    def write_to_stdout(self, command):
        print "NOT CONNECTED: Sending: ", command

    def write_to_printer(self, command):
        print "Sending: ", command
        res = 'ok'
        while True:
            if not 'wait' in res:
                self.printer.write(command.strip()+"\n")

            res = self.printer.readline()
            while not res:
                res = self.printer.readline()
            print "Response: ", res
            if "ok" in res:
                break

    def generate_circumference(self,n=100):
        points = sorted(list(set([(round(math.cos(2*pi/n*x)*self.r), round(math.sin(2*pi/n*x)*self.r)) for x in xrange(0,n+1)])))
        self.redraw(self.axes.scatter([x[0] for x in points], [y[1] for y in points], marker="+", color='red'))

    def generate_coordinates(self):
        for y in np.arange(self.r, -self.r, -self.step_size):
            for x in np.arange(-self.r, self.r, self.step_size):
                if math.sqrt(x**2+y**2) <= self.r:
                    self.points.append(('inside', x, y))
                    self.redraw(self.axes.scatter(x, y, color='blue'))

                else:
                    self.points.append(('outside', x, y))
                    self.redraw(self.axes.scatter(x, y, marker=".", color='black'))

    def redraw(self, points):
        self.axes.draw_artist(points)
        self.figure.canvas.blit(self.bbox)

    def main(self):
        self.generate_circumference()
        self.generate_coordinates()
        self.connect_to_printer()
        self.configure()

    def configure(self):
        cols = (self.r*2)/self.step_size
        rows = (self.r*2)/self.step_size
        current_col = 0
        current_row = 0

        row_change_point = False

        self.writer("G28")
        time.sleep(self.wth)
        self.writer("G1 X0 Y0 Z0 F{0}".format(self.travel))
        time.sleep(self.wtz)

        with open(self.csv_file, 'w') as f:
            for where, x, y in self.points:

                if where == 'inside':

                    if row_change_point:
                        row_change_point = False
                        self.writer("G1 X{0} Y{1} Z{2} F{3}".format(x, y, self.z_height_row_change, self.travel))
                        time.sleep(self.wtp)

                    self.modify_point(x, y)
                    gcode = "G1 X{0} Y{1} Z0 F{2}".format(x, y, self.point_speed)
                    self.writer(gcode)
                    measurement = self.mf(x, y)
                    f.write("{0}".format(measurement))
                    self.points_adjusted.append((x, y, measurement))
                    time.sleep(self.wtp)
                else:
                    self.modify_point(x, y, marker='.', color='yellow')
                    measurement = 0.00
                    f.write(str(measurement))

                current_col += 1
                print "Row: {0} of {1} - COL {2} OF {3} (Z: {4})".format(current_row, rows, current_col, cols, measurement)

                if current_col == cols:
                    current_col = 0
                    current_row += 1
                    f.write("\n")
                    self.writer("G1 Z{0} F{1}".format(self.z_height_row_change, self.travel))
                    time.sleep(self.wtr)
                    row_change_point = True
                else:
                    f.write(",")

    def modify_point(self, x, y, marker='o', color='orange'):
        self.redraw(self.axes.scatter(x, y, marker=marker, color=color))

    def generate_buildplate_image(self):
        plt.close()
        plt.ioff()

        from mpl_toolkits.mplot3d import Axes3D
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        x = [x[0] for x in self.points_adjusted]
        y = [y[1] for y in self.points_adjusted]
        z = [z[2] for z in self.points_adjusted]

        ax.set_ylim(-self.r-5, self.r+5)
        ax.set_xlim(-self.r-5, self.r+5)
        ax.set_zlim(0, 3*max(z))

        ax.scatter(x, y, z, zdir=u'z', s=10, c=u'b', depthshade=True)
        plt.show()